use anyhow::Result;
use std::{collections::HashMap, fs::File, io::Read, num::NonZeroU32};
use wgpu::BindGroupLayout;

use super::rawtexture::RawTexture;

pub struct TextureArray {
	pub bind_group: wgpu::BindGroup,
	pub bind_group_layout: BindGroupLayout,
	texture_dims: Vec<(u32, u32)>,
	texture_indices: HashMap<String, u32>,
}

const TEX: &[&str] = &["error", "red", "gypsum", "earth", "a0"];

pub struct Texture {
	pub index: u32,  // Index in texture array
	pub aspect: f32, // width / height
	pub height: f32, // Height in game units
}

impl TextureArray {
	pub fn get_texture(&self, name: &str) -> Texture {
		let index = match self.texture_indices.get(name) {
			Some(x) => *x,
			None => 0, // Default texture
		};

		let dimensions = self.texture_dims[index as usize];

		return Texture {
			index,
			height: 100.0,
			aspect: dimensions.0 as f32 / dimensions.1 as f32,
		};
	}

	pub fn new(device: &wgpu::Device, queue: &wgpu::Queue) -> Result<Self> {
		let mut textures: Vec<RawTexture> = Vec::new();
		let mut texture_indices: HashMap<String, u32> = HashMap::new();

		let mut i = 0;
		for t in TEX {
			let p = format!("assets/{t}.png");
			let mut f = File::open(&p)?;
			let mut bytes = Vec::new();
			f.read_to_end(&mut bytes)?;
			textures.push(RawTexture::from_bytes(&device, &queue, &bytes, &p).unwrap());
			texture_indices.insert(t.to_string(), i);
			i += 1;
		}

		// One of these overall
		let sampler = device.create_sampler(&wgpu::SamplerDescriptor {
			address_mode_u: wgpu::AddressMode::ClampToEdge,
			address_mode_v: wgpu::AddressMode::ClampToEdge,
			address_mode_w: wgpu::AddressMode::ClampToEdge,
			mag_filter: wgpu::FilterMode::Linear,
			min_filter: wgpu::FilterMode::Nearest,
			mipmap_filter: wgpu::FilterMode::Nearest,
			..Default::default()
		});

		let bind_group_layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
			label: Some("texture_bind_group_layout"),
			entries: &[
				// Texture data
				wgpu::BindGroupLayoutEntry {
					binding: 0,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Texture {
						multisampled: false,
						view_dimension: wgpu::TextureViewDimension::D2,
						sample_type: wgpu::TextureSampleType::Float { filterable: true },
					},
					count: NonZeroU32::new(textures.len() as u32),
				},
				// Texture sampler
				wgpu::BindGroupLayoutEntry {
					binding: 1,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
					count: NonZeroU32::new(textures.len() as u32),
				},
			],
		});

		let views: Vec<&wgpu::TextureView> = textures.iter().map(|x| &x.view).collect();
		let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
			label: Some("texture_bind_group"),
			layout: &bind_group_layout,
			entries: &[
				wgpu::BindGroupEntry {
					binding: 0,
					// Array of all views
					resource: wgpu::BindingResource::TextureViewArray(&views),
				},
				wgpu::BindGroupEntry {
					binding: 1,
					resource: wgpu::BindingResource::SamplerArray(&[&sampler].repeat(views.len())),
				},
			],
		});

		return Ok(Self {
			bind_group,
			bind_group_layout,
			texture_dims: textures.iter().map(|x| x.dimensions).collect(),
			texture_indices,
		});
	}
}
